/*
 * packetmodeller/init.c
 *
 * Copyright (C) 2008-2009 Reto Gantenbein
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>
#include "init.h"
#include "logger.h"
#include "message.h"
#include "netdevice.h"
#include "packet_modeller.h"
#include "typedef.h"

/** Declare local wait indicator */
int wait;


/** Read a port from char array
 *  check for validity and return it in network byte order */
uint16_t
check_port(char *cport)
{
    uint16_t iport = atoi(cport);
    if (iport > 0) // don't need to check upper limit becaues of uint16_t range
    {
        return htons(iport);
    } else {
        printf("Error: Invalid port given\n");
        return 0;
    }
}

/** Checks char array for valid IP address (0.0.0.0 to 255.255.255.255) */
int
check_ip(char *ip)
{
    int err = 0;
    unsigned b1, b2, b3, b4;

    if (sscanf(ip, "%3u.%3u.%3u.%3u", &b1, &b2, &b3, &b4) != 4)
        err = 1;
    if (b1 > 255 || b2 > 255 || b3 > 255 || b4 > 255)
        err = 1;
    if (err)
        printf("Error: Invalid model-ip address given\n");

    return err;
}

/** Checks IP:Port pair for validity and fills into sockaddr_in struct */
int
check_model_addr(struct sockaddr_in *addr, char *chost)
{
    int ip_char;
    const char delimiter = ':';

    for (ip_char = 0; ip_char < 16; ip_char++)
    {
        if (*(chost + ip_char) == delimiter)
        {
            char ip_array[ip_char + 1];
            memcpy(ip_array, chost, ip_char);
            ip_array[ip_char] = '\0';
            if (check_ip(ip_array) != 0)
                return -1;
            inet_aton(ip_array, &((*addr).sin_addr));
            if (((*addr).sin_port = check_port(chost + ip_char + 1)) == 0)
                return -1;
            (*addr).sin_family = AF_INET;
            return 0;
        }
    }
    printf("Error: Invalid format model-ip:model-port given\n");

    return -1;
}

/** Check interface name for availability and adds it to the interfaces array */
int
check_if(int if_index, char *if_name)
{
    ifname iface;
    char *ifp;
    int if_len = strlen(if_name);

    ifp = (char *) iface;
    strncpy(ifp, (char *) if_name, if_len);
    ifp[if_len] = '\0';
    if (get_ip_by_dev(NULL, (const ifname *) iface) == -1)
    {
        printf("Error: Invalid interface %s\n", if_name);
        return -1;
    }
    strncpy(interfaces[if_index], iface, IF_NAMESIZE);
    print_dbg("registering interfaces[%i]=%s\n", \
        if_index, interfaces[if_index]);

    return 0;
}

/** Fill the arg_array with pointer to the comma seperated arguments
 *  and returns the number of arguments (1 or 2) */
int
get_args(char *arg_array[], char *args)
{
    int num_arg = 1;
    arg_array[0] = args;
    const char delimiter = ',';
    char *second_arg = strchr(args, delimiter);

    if (second_arg != NULL)
    {
        *second_arg = '\0';
        second_arg++;
        num_arg++;
    }
    arg_array[1] = second_arg;

    return num_arg;
}

/** Parse command line options and checks it for validity */
void
pars_opts(int argc, char *argv[])
{
    int opt;
    int req_args = 3;
    int add_args = 0; // required because getopts changes the order of arguments
    int ret_args = 0;
    char *arg_array[2];
    debug = 0;
    direct = 0;

    while ((opt = getopt(argc, argv, "dnhp:w:")) != -1)
    {
        switch (opt)
        {
            case 'h':
                print_help();
                break;
            case 'p':
                add_args += 2;
                break;
            case 'w':
                add_args += 2;
                break;
            case 'd':
                add_args++;
                debug = 1;
                break;
            case 'n':
                add_args++;
                break;
        }
    }
    optind = 0; // reset getopt execution

    if (argc < (req_args + add_args))
    {
        printf("Error: Missing arguments\n");
        print_usage();
    }

    if (debug == 1)
    {
        print_dbg = printf;
        dbg_msg = dump_msg;
    } else {
        print_dbg = ignore_dbg;
        dbg_msg = ignore_dbg;
    }
    print_dbg("-> debugging enabled\n");

    // read model address
    if (check_model_addr(&model_addr, argv[1 + add_args]) == -1)
        print_usage();

    // read local port
    if ((node_port = check_port(argv[2 + add_args])) == 0)
        print_usage();

    while ((opt = getopt(argc, argv, "dnp:w:")) != -1)
    {
        switch (opt)
        {
            case 'n':
                direct = 1;
                break;
            case 'p':
                if (check_if(0, optarg))
                    print_usage();
                break;
            case 'w':
                ret_args = get_args(arg_array, optarg);
                for (opt = 0; opt < ret_args; opt++)
                {
                    if (check_if(opt + 1, arg_array[opt]))
                        print_usage();
                }
                break;
        }
    }
}

/** Clears wait status */
void
clear_wait()
{
    wait = 0;
}

/** Register the host at the WlanModel */
void
register_node()
{
    if (direct == 1)
    {
        print_dbg("node-to-node communication enabled: ");
        print_dbg("not registering node!\n");
        return;
    } else {
        struct register_msg *reg_msg;
        char reg_buffer[REGISTER_MSG_SIZE];
        char ack_buffer[NODE_PREAMBLE_SIZE + MAX_MSG_SIZE];
        uint16_t message_id = 0;
        int recv_bytes = 0;
        uint8_t success = 0;

        reg_msg = (struct register_msg *) reg_buffer;
        (*reg_msg).header.type = REGISTER_MSG;
        memmove(&(*reg_msg).header.id, node_id, sizeof(host_id));
        memmove(&(*reg_msg).name, host_name, HOSTNAME_SIZE);
        memmove(&(*reg_msg).ip, &node_ip, sizeof(struct in_addr));
        reg_msg->port = node_port;
        signal(SIGALRM, clear_wait);

        while (success == 0)
        {
            (*reg_msg).msg_id = message_id;
            wait = 1;
            dbg_msg((char *) reg_msg, 0);

            // send register message to model
            if (sendto(net_socket, reg_msg, REGISTER_MSG_SIZE, 0, \
                (struct sockaddr *) &model_addr, sizeof(struct sockaddr_in))
                < 0)
            {
                printf("register_node(): Sending register_msg failed (%s)\n", \
                    strerror(errno));
            }
            print_dbg("[net_socket] %i bytes sent (%s)\n", REGISTER_MSG_SIZE,
                    strerror(errno));

            // wait 5s for ack message else resend register message
            alarm(5);
            while (wait == 1)
            {
                if ((recv_bytes = recv(net_socket, ack_buffer, \
                    sizeof(ack_buffer), MSG_DONTWAIT)) > 0)
                {
                    print_dbg("[net_socket] %i bytes read (%s)\n", \
                        recv_bytes, strerror(errno));
                    dbg_msg(ack_buffer, 0);
                    if (isAcknowledgeMessage((struct msg_header *) ack_buffer) \
                        &&
                        (((struct ack_msg *) ack_buffer)->msg_id <= message_id))
                    {
                        print_dbg("node successfully registered!\n");
                        alarm(0);
                        wait = 0;
                        success = 1;
                    }
                } else {
                    errno = 0;
                }
            }
            message_id++;
        }
    }
}


/** Initialize network file descriptors */
void
init_fds()
{
    // initialize tap device
    print_dbg("opening %s socket...", (char *) interfaces[1]);
    if ((tap_fd = alloc_tap_socket((const ifname *) &interfaces[1])) > 1)
    {
        print_dbg("OK (tap_fd=%i)\n", tap_fd);
    } else {
        exit(1);
    }

    // initialize network device
    print_dbg("opening %s socket...", (char *) interfaces[0]);
    if ((net_socket = alloc_net_socket((const ifname *) &interfaces[0], \
        ntohs(node_port))) > 1)
    {
        print_dbg("OK (net_socket=%i)\n", net_socket);
    } else {
        exit(1);
    }
}


/** Initialize node variables */
int
init_node()
{
    int i;
    mac_addr mac;

    if (gethostname(host_name, HOSTNAME_SIZE) != 0)
    {
        printf("%s:%i - Unable to retrieve host name (%s)\n", \
            __FILE__, __LINE__, strerror(errno));
        return -1;
    }

    // create host id from physical mac address
    get_mac_by_dev(mac, (const ifname *) &interfaces[0]);
    for (i = 0; i < sizeof(mac); ++i)
    {
        sprintf((char *) node_id + 2*i, "%2.2x", ((char *)mac)[i]);
    }
    node_id[sizeof(node_id)] = "\0";

    if (get_ip_by_dev(&node_ip, (const ifname *) &interfaces[0]) != 0)
    {
        printf("Error: Invalid interface %s\n", interfaces[0]);
        return -1;
    }

    printf("****************************************\n");
    printf("Node Name: %s\n", host_name);
    printf("Node ID: %s\n", node_id);
    printf("Node IP: %s\n", inet_ntoa(node_ip));
    printf("Node Port: %i\n", ntohs(node_port));
    printf("****************************************\n");

    return 0;
}
