/*
 * wlanmodel/packetproxy.cc
 *
 * Copyright (C) 2008-2009 Reto Gantenbein
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <cstdlib>
#include <string>
#include <omnetpp.h>
#include <arpa/inet.h>
#include <Ieee802Ctrl_m.h>
#include "crawsocketrtscheduler.h"
#include "logger.h"
#include "message.h"
#include "nodeinfo.h"
#include "nodemanager.h"
#include "packetproxy.h"
#include "rawetherframe_m.h"
#include "typedef.h"

/** Register as simulation module */
Define_Module(PacketProxy);

/**
 * Constructor
 */
PacketProxy::PacketProxy()
{
    rtEvent = NULL;
}

/**
 * Destructor
 */
PacketProxy::~PacketProxy()
{
    cancelAndDelete(rtEvent);
}

/**
 * Initialize PacketProxy at simulation startup
 */
void
PacketProxy::initialize()
{
    rtEvent = new cMessage("rtEvent");
    rtScheduler = check_and_cast<cRAWSocketRTScheduler *>(simulation.scheduler());
    rtScheduler->setInterfaceModule(this, rtEvent, inBuffer, \
        NODE_PREAMBLE_SIZE + MAX_MSG_SIZE, &numRecvBytes);
    pktProcTime.setName("Packet Processing Time");
}

/**
 * Handle Omnet++ messages
 */
void
PacketProxy::handleMessage(cMessage *msg)
{
    ev << "PacketProxy.handleMessage(): " << msg << " received\n";
    if (msg == rtEvent)
    {
        handleSocketEvent(msg);
    } else {
        handleReply(check_and_cast<RAWEtherFrame *> (msg));
    }
}

/**
 * Process packets which are received from the network socket
 */
void
PacketProxy::handleSocketEvent(cMessage *rtMsg)
{
    switch(getMessageType((msg_header *) inBuffer))
    {
        case NODE_MSG:
            ev << "PacketProxy.handleSocketEvent(): NODE_MSG received\n";
            processNodeMsg(rtMsg);
            break;
        case REGISTER_MSG:
            ev << "PacketProxy.handleSocketEvent(): REGISTER_MSG received\n";
            registerNode(rtMsg);
            break;
        default:
            ev << "PacketProxy.handleSocketEvent(): Unknown message received ("\
                << getMessageType((struct msg_header*) inBuffer) << ")\n";
            numRecvBytes = 0;
            break;
    }
}

/**
 * Process packets which are received from the wireless simulation
 */
void
PacketProxy::handleReply(RAWEtherFrame *etherFrame)
{
    int i;
    struct sockaddr_in *nodeAddr;
    NodeInfo *destinationHost;
    int arrivalGateIndex = etherFrame->arrivalGate()->index();
    int payloadSize = etherFrame->getRawBytesArraySize();
    int messageSize = MODEL_PREAMBLE_SIZE + payloadSize;
    model_msg *modelMsg = (struct model_msg *) calloc(1, messageSize);

    ((struct msg_header*) modelMsg)->type = MODEL_MSG;
//  ((struct msg_header*) modelMsg)->id = ????
    char *payloadPointer = (char *) &(modelMsg->wlan_pkt);
    for (i = 0; i < payloadSize; i++)
    {
        *payloadPointer = etherFrame->getRawBytes(i);
        payloadPointer++;
    }
    nodeAddr = (struct sockaddr_in*) calloc(1, sizeof(struct sockaddr_in));
    if ((destinationHost = NodeManager::getNodeInfo(arrivalGateIndex)) != NULL)
    {
        destinationHost->getNodeAddr(nodeAddr);
        ev << "PacketProxy.handleReply(): target host = " << \
            destinationHost->getHostId();
        ev << "(" << destinationHost->getHostName();
        in_addr *ipAddr = (in_addr *) &(nodeAddr->sin_addr);
        ev << ", " << inet_ntoa(*ipAddr) << ")\n";
        rtScheduler->sendMessage((const char*) modelMsg, messageSize, nodeAddr);

        timeval curTime;
        gettimeofday(&curTime, NULL);
        timeval processingTimeval = \
            timeval_substract(curTime, etherFrame->timestamp());
        double processingTime = (processingTimeval.tv_sec + \
            processingTimeval.tv_usec * 1e-6) * 1000;  // in ms
        ev << "PacketProxy.handleReply(): processingTime " << \
            processingTime << "ms\n";
        pktProcTime.record(processingTime);
    } else {
        ev << "PacketProxy.handleReply(): ERROR - unknown target host\n";
    }

    delete etherFrame;
    free(modelMsg);
    free(nodeAddr);
}

/**
 * Check if host is registered and notify ProxyHost of incoming NodeMsg
 */
void
PacketProxy::processNodeMsg(cMessage *rtMsg)
{
    const host_id *nodeId = (host_id*) ((msg_header*) inBuffer)->id;

    if (!NodeManager::isRegistered(nodeId)) {
        numRecvBytes = 0;                   //discard received packet

    } else {
        int gateIndex = NodeManager::getProxyGate(nodeId);
        rtMsg->setKind(gateIndex);           //set output gate
        sendPacket(rtMsg);
        numRecvBytes = 0;
    }
}

/**
 * Sends received network packets to simulation wireless stack
 */
void
PacketProxy::sendPacket(cMessage *msg)
{
    int i;
    node_msg *nodeMsg = (node_msg *) inBuffer;

    //set wireless information
    int channel = nodeMsg->channel;
    int transmitPower = nodeMsg->transmit_power;
    RAWEtherFrame *etherFrame = new RAWEtherFrame("etherFrame");

    etherFrame->setTimestamp(msg->timestamp());
    etherFrame->setLength(nodeMsg->pkt_length);
    etherFrame->setRawBytesArraySize(nodeMsg->pkt_length);
    char *bufferPtr = (char *) &(nodeMsg->wlan_pkt);
    for (i = 0; i < (nodeMsg->pkt_length); i++)
    {
        etherFrame->setRawBytes(i, *bufferPtr);
        bufferPtr++;
    }
    MACAddress destMacAddr = getDestMacAddr((RAWEtherFrame *) etherFrame);
    Ieee802Ctrl *controlInfo = new Ieee802Ctrl();
    controlInfo->setDest(destMacAddr);
    etherFrame->setControlInfo(controlInfo);

    ev << "PacketProxy.sendPacket(): sending EtherFrame to " \
        << destMacAddr.str() << "\n";
    send(etherFrame, "send", msg->kind());
}

/**
 * Register node at the NodeManager
 */
void
PacketProxy::registerNode(cMessage *rtMsg)
{
    std::string hostName;
    register_msg *regMsg = (register_msg *) inBuffer;
    host_id *hostId = (host_id *) &((regMsg->header).id);
    uint16_t msgId = regMsg->msg_id;
    const char *charName = (const char *) &(regMsg->name);

    hostName.insert(0, charName);
    in_addr *hostIp = &(regMsg->ip);
    uint16_t port = regMsg->port;
    NodeInfo *info = new NodeInfo(hostId, hostName, hostIp, port);
    if (! NodeManager::isRegistered(hostId))
    {
        NodeManager::addNode(info);
    }
    numRecvBytes = 0;
    sendAcknowledgeMsg(msgId, info);
}

/**
 * Send acknowledge message to the node
 */
int
PacketProxy::sendAcknowledgeMsg(uint16_t msg_id, NodeInfo *node)
{
    ack_msg *ack;
    sockaddr_in *nodeAddr;

    ack = (ack_msg*) calloc(1, ACK_MSG_SIZE);
    nodeAddr = (sockaddr_in*) calloc(1, sizeof(sockaddr_in));
    ack->msg_id = msg_id;
    ((msg_header*) ack)->type = ACK_MSG;
    node->getNodeAddr(nodeAddr);
    in_addr *ipAddr = (in_addr *) &(nodeAddr->sin_addr);
    ev << "PacketProxy.sendAcknowledgeMsg(): to " << node->getHostName() << \
        " (" << inet_ntoa(*ipAddr) << "), "<< "id: " << msg_id << "\n";
    int err = rtScheduler->sendMessage((const char*) ack, ACK_MSG_SIZE, nodeAddr);

    free(ack);
    free(nodeAddr);
    return err;
}

/**
 * Return message type
 */
unsigned char
PacketProxy::getMessageType(struct msg_header *header)
{
    return header->type;
}

/**
 * Extracts the destination MAC address from a RAWEtherFrame
 * --> TODO: move to RAWEtherFrame class
 */
MACAddress
PacketProxy::getDestMacAddr(RAWEtherFrame *msg)
{
    int i;
    mac_addr destMac;
    char *charPtr = (char *) destMac;

    for (i = 0; i < 6; i++)
    {
        *charPtr = msg->getRawBytes(i);
        charPtr++;
    }

    MACAddress macAddr;
    macAddr.setAddressBytes((unsigned char *) destMac);
    return macAddr;
}
